# Copyright (c) 2003 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

"""
Unit tests for aspects
"""

__revision__ = '$Id: unittest_contracts.py,v 1.4 2003/06/17 09:24:10 alf Exp $'


import unittest

from logilab.aspects.weaver import weaver
from logilab.aspects.lib.contracts import ContractAspect, ContractFailedError



class Foo:
    """
    inv:
        self.invariant == 12
    """

    def __init__(self):
        self.invariant = 12
    
    def silly_func(self, value, *args, **kwargs):
        """
        pre:
            value < 10
        post:
            type(__return__) == type('') and int(__return__) < 15
            # Silly condition (just for tests on __old__)
            __old__.value != __old__.self.invariant
        """
        return "%d"%value
    

    def bad_post(self, val, *args, **kwargs):
        """
        pre:
            val < 10
        post:
            type(__return__) == type('') and int(__return__) < 15
        """
        
        return "20"


    def sort(self, array):
        """
        post:
            # array size is unchanged
            len(array) == len(__old__.array)
        
            # array is ordered
            forall([array[i] >= array[i-1] for i in range(1, len(array))])
        """
        array.sort()


    def corrupt(self):
        """This method will raise an exception after changing
        invariant to check that invariant are still checked even
        if an exception was raised
        """
        self.invariant = 14
        raise Exception, "..."


    def post_strength(self):
        """
        post:
            __return__ < 10
        """
        return 5

    
    
class Bar(Foo):
    """Just a class to test inheritance csq in contracts
    """

    def __init__(self):
        Foo.__init__(self)
        self.age = 12
        
    def silly_func(self, value, *args, **kwargs):
        """
        pre:
            True
        post:
            self.age == __old__.self.age * 2
        """        
        self.age *= 2
        return "%d"%value


    def bad_post(self, val, *args, **kwargsl):
        """ Always bad post condition !
        """
        return 20


    def post_strength(self):
        """
        post
            __return__ > 5
        """
        return 5


class Contract_ClassTest(unittest.TestCase):
    """TestCase for aspects
    """

    def setUp(self):
        """setUp method for tests
        """
        self.a = Foo()
        self.b = Bar()
        weaver.weave_methods(self.a, ContractAspect)
        weaver.weave_methods(self.b, ContractAspect)
        self.aspect = ContractAspect(Bar,'silly_func')
        
    def test_pre_conditions(self):
        """Tests if preconditions work
        """
        self.a.silly_func(3)
        self.assertRaises(ContractFailedError, self.a.silly_func, 20)
    

    def test_post_conditions(self):
        """Tests if postconditions work
        """
        bar_post = ['self.age == getattr(__old__,"self.age") * 2']
        foo_post = ["type(__return__) == type('') and int(__return__) < 15",
                    'getattr(__old__,"value") != ' \
                    'getattr(__old__,"self.invariant")']
        self.assert_(self.aspect.post == [bar_post, foo_post])
        self.assertRaises(ContractFailedError, self.a.bad_post,5)
        
        
    def test_inheritance(self):
        """Tests inheritance
        """
        self.assert_(self.aspect.pre == [['True'], ['value < 10']])

    def test_preconditions_weaken(self):
        """Tests if pre_conditions are weaken
        """
        # If this test passes, then the pre-condition has been weaken
        self.b.silly_func(13)



    def test_postconditions_strengthen(self):
        """Tests if post_conditions are strengthen
        """
        self.a.post_strength()
        # self.assertRaises(ContractFailedError, self.b.post_strength)
        
        

    def test_invariant(self):
        """Tests if invariant conditions work
        """
        
        self.assert_(self.aspect.inv == [['self.invariant == 12']])
        # This test should pass
        self.a.silly_func(3)
        self.a.invariant = 15
        # This one should not pass since the invariant condition is not
        # true anymore
        self.assertRaises(ContractFailedError, self.a.silly_func, 3)


    def test_old_variable(self):
        """Tests if __old__ works well
        """
        self.assert_(self.aspect.memo_list == ['self.age', 'value',
                                               'self.invariant'])
        

    def test_forall_func(self):
        """Tests if forall works well
        """
        self.a.sort([12,5,8,19,0])


    def test_exception_process(self):
        """Checks if exception are correctly processed
        """
        try:
            self.a.corrupt()
        except ContractFailedError, error:
            self.assert_(error.method_exception is not None)
        else:
            self.fail("No invariant checks after exception raised !")


    def test_func_dict(self):
        """Checks if the wrapped func's global_dict is
        corretcly taken in account.
        """
        import logilab.aspects.tests.module_test
        weaver.weave_methods(logilab.aspects.tests.module_test.Sorter,
                             ContractAspect)
        sorter = logilab.aspects.tests.module_test.Sorter()
        sorter.sort([12,1,5])
        self.assertRaises(ContractFailedError, sorter.sort, "bad input")
        

    

def suite():
    """Callable object that returns the entire testsuite
    (useful for alltests.py)
    """
    return unittest.makeSuite(Contract_ClassTest,"test")


if __name__ == '__main__':
    unittest.main()

