# Copyright (c) 2003 LOGILAB S.A. (Paris, FRANCE).
# http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

"""
Unit tests for Weaver
"""

__revision__ = '$Id: unittest_weaver.py,v 1.12 2003/06/17 09:24:11 alf Exp $'


import unittest

from logilab.aspects.weaver import weaver
from logilab.aspects.lib.contracts import ContractAspect, ContractFailedError
from logilab.aspects.lib.logger import LoggerAspect
import logilab.aspects.tests.module_test

import sys

class Test:
    """Small class used in Weaver_ClassTest
    """
    def foo(self, val):
        """
        pre:
            val < 10
        """
        return val


    def bar(self):
        """ nothing to do ...
        """
        pass


class Weaver_ClassTest(unittest.TestCase):
    """TestCase for the aspect weaver.
    """

    def setUp(self):
        """setUp method for tests
        """
        self.instance = Test()
        self.klass    = Test
        self.log_device = sys.stdout
        self.mod = logilab.aspects.tests.module_test
        

    def test_weave_methods(self):
        """Tests if the weaver weaves correctly methods
        """
        old_foo = getattr(self.instance, 'foo')
        old_bar = getattr(self.instance, 'bar')
        weaver.weave_methods(self.instance, ContractAspect)
        self.assert_(self.instance in weaver._Weaver__woven_dict)
        self.assert_(self.klass not in weaver._Weaver__woven_dict)
        new_foo = getattr(self.instance, 'foo')
        new_bar = getattr(self.instance, 'bar')
        self.assert_(old_foo != new_foo)
        self.assert_(old_bar != new_bar)


    def test_weave_instance(self):
        """Tests if only the right instance is weaved
        """
        weaver.weave_methods(self.instance, ContractAspect)
        self.assertRaises(ContractFailedError, self.instance.foo, 15)
        another_instance = Test()
        # In another instance : contracts are not activated, this should pass !
        ret_v = another_instance.foo(15)
        # Clean up
        weaver.unweave_methods(self.instance, ContractAspect)
        

    def test_weave_class(self):
        """Tests if all new instances of a class are also weaved
        """
        weaver.weave_methods(self.klass, ContractAspect)
        self.assertRaises(ContractFailedError, self.instance.foo, 15)
        another_instance = Test()
        self.assertRaises(ContractFailedError, another_instance.foo, 15)
        # Clean up
        weaver.unweave_methods(self.klass, ContractAspect)


    def test_unweave_methods(self):
        """Tests if the weaver unweaves correctly methods
        """
        weaver.weave_methods(self.klass, ContractAspect)
        self.assertRaises(ContractFailedError, self.instance.foo, 15)
        weaver.unweave_methods(self.klass, ContractAspect)
        # Contracts are no more activated : this should pass !
        ret_v = self.instance.foo(15)
        # Test on instances
        another_instance = Test()
        weaver.weave_methods(self.instance, ContractAspect)
        weaver.weave_methods(another_instance, ContractAspect)
        self.assertRaises(ContractFailedError, self.instance.foo, 15)
        self.assertRaises(ContractFailedError, another_instance.foo, 15)
        weaver.unweave_methods(self.instance, ContractAspect)
        # Contracts are no more activated on self.instance: this should pass !
        ret_v = self.instance.foo(15)
        # They should be still activated on another_instance
        self.assertRaises(ContractFailedError, another_instance.foo, 15)
        

    def test_multiple_weaving(self):
        """Tests if multiple aspects can be (un)weaved
        """
        log_device = open('tests.log','w')
        weaver.weave_methods(self.instance, ContractAspect)
        weaver.weave_methods(self.instance, LoggerAspect, log_device)
        self.assert_(len(weaver.get_aspects(self.instance,
                                                 'foo')) == 2)
        weaver.unweave_methods(self.instance, ContractAspect)
        self.assert_(len(weaver.get_aspects(self.instance,
                                                 'foo')) == 1)
        # Tests if the remaining Aspect is not the Contract one.
        ret_v = self.instance.foo(15)
        log_device.close()


    def test_weave_module(self):
        """Tests if module's weaving works correctly
        """
        weaver.weave_module(self.mod, ContractAspect)
        stack = self.mod.StackImpl(3)
        self.assertRaises(ContractFailedError, stack.pop)
        # Clean up
        weaver.unweave_module(self.mod, ContractAspect)


    def test_unweave_module(self):
        """Tests if module's unweaving works correctly
        """
        weaver.weave_module(self.mod, ContractAspect)
        weaver.unweave_module(self.mod, ContractAspect)
        # Create a stack of max_size = 1 and push 2 elements
        # if module has been unweaved, this should be ok !
        stack = self.mod.StackImpl(1)
        stack.push(1)
        stack.push(2)


    def test_weave_same_aspect(self):
        """Checks that weaver forbids weaving multiple time the same
        aspect on a method.
        """
        weaver.weave_methods(self.klass, ContractAspect)
        old_aspects = weaver.get_aspects(self.klass, 'foo')
        weaver.weave_methods(self.klass, ContractAspect)
        new_aspects = weaver.get_aspects(self.klass, 'foo')
        self.assert_(old_aspects == new_aspects)
        
    
def suite():
    """Callable object that returns the entire testsuite
    (useful for alltests.py)
    """
    return unittest.makeSuite(Weaver_ClassTest,"test")



if __name__ == '__main__':
    unittest.main()

